import numpy as np


class _CheckInputs:
    """Check if additional arguments are correct"""

    def __init__(self, n):
        self.n = n

    def __call__(self, **kwargs):
        if type(self.n) is not int:
            raise ValueError("Expected n of type int, got {}".format(type(self.n)))

        if self.n < 5:
            raise ValueError("n must be >= 5, got {}".format(self.n))

        for key, value in kwargs.items():
            if value[1] is float and type(value[0]) is int:
                continue
            if type(value[0]) is not value[1]:
                raise ValueError(
                    "Expected {} type {} got {}".format(key, value[1], type(value[0]))
                )


def indep_ar(n, lag=1, phi=0.5, sigma=1):
    r"""
    2 independent, stationary, autoregressive time series simulation.

    :math:`X_t` and :math:`Y_t` are univarite AR(``1ag``) with
    :math:`\phi = 0.5` for both series. Noise follows
    :math:`\mathcal{N}(0, \sigma)`. With lag (1), this is

    .. math::

        \begin{bmatrix} X_t \\ Y_t \end{bmatrix} =
        \begin{bmatrix} \phi & 0 \\ 0 & \phi \end{bmatrix}
        \begin{bmatrix} X_{t - 1} \\ Y_{t - 1} \end{bmatrix} +
        \begin{bmatrix} \epsilon_t \\ \eta_t \end{bmatrix}

    Parameters
    ----------
    n : int
        The number of samples desired by the simulation (>= 3).
    lag : float, default: 1
        The maximum time lag considered between ``x`` and ``y``.
    phi : float, default: 0.5
        The AR coefficient.
    sigma : float, default: 1
        The variance of the noise.

    Returns
    -------
    x,y : ndarray of float
        Simulated data matrices. ``x`` and ``y`` have shape ``(n,)``
        where `n` is the number of samples.
    """
    extra_args = {
        "lag": (lag, float),
        "phi": (phi, float),
        "sigma": (sigma, float),
    }
    check_in = _CheckInputs(n)
    check_in(**extra_args)

    epsilons = np.random.normal(0, sigma, n)
    etas = np.random.normal(0, sigma, n)

    x = epsilons
    y = etas

    # AR process
    for t in range(lag, n):
        x[t] = phi * x[t - lag] + epsilons[t]
        y[t] = phi * y[t - lag] + etas[t]

    return x, y


def cross_corr_ar(n, lag=1, phi=0.5, sigma=1):
    r"""
    2 linearly dependent time series simulation.

    :math:`X_t` and :math:`Y_t` are together a bivariate univarite AR(``1ag``) with
    :math:`\phi = \begin{bmatrix} 0 & 0.5 \\ 0.5 & 0 \end{bmatrix}` for both series.
    Noise follows :math:`\mathcal{N}(0, \sigma)`. With lag (1), this is

    .. math::

        \begin{bmatrix} X_t \\ Y_t \end{bmatrix} =
        \begin{bmatrix} 0 & \phi \\ \phi & 0 \end{bmatrix}
        \begin{bmatrix} X_{t - 1} \\ Y_{t - 1} \end{bmatrix} +
        \begin{bmatrix} \epsilon_t \\ \eta_t \end{bmatrix}

    Parameters
    ----------
    n : int
        The number of samples desired by the simulation (>= 3).
    lag : float, default: 1
        The maximum time lag considered between ``x`` and ``y``.
    phi : float, default: 0.5
        The AR coefficient.
    sigma : float, default: 1
        The variance of the noise.

    Returns
    -------
    x,y : ndarray of float
        Simulated data matrices. ``x`` and ``y`` have shape ``(n,)``
        where `n` is the number of samples.
    """
    extra_args = {
        "lag": (lag, float),
        "phi": (phi, float),
        "sigma": (sigma, float),
    }
    check_in = _CheckInputs(n)
    check_in(**extra_args)

    epsilons = np.random.normal(0, sigma, n)
    etas = np.random.normal(0, sigma, n)

    x = epsilons
    y = etas

    for t in range(lag, n):
        x[t] = phi * y[t - lag] + epsilons[t]
        y[t] = phi * x[t - lag] + etas[t]

    return x, y


def nonlinear_process(n, lag=1, phi=1, sigma=1):
    r"""
    2 nonlinearly dependent time series simulation.

    :math:`X_t` and :math:`Y_t` are together a bivariate nonlinear process.
    Noise follows :math:`\mathcal{N}(0, \sigma)`. With lag (1), this is

    .. math::

        \begin{bmatrix} X_t \\ Y_t \end{bmatrix} =
        \begin{bmatrix} \phi \epsilon_t Y_{t - 1} \\ \eta_t \end{bmatrix}

    Parameters
    ----------
    n : int
        The number of samples desired by the simulation (>= 3).
    lag : float, default: 1
        The maximum time lag considered between `x` and `y`.
    phi : float, default: 1
        The AR coefficient.
    sigma : float, default: 1
        The variance of the noise.

    Returns
    -------
    x,y : ndarray of float
        Simulated data matrices. ``x`` and ``y`` have shape ``(n,)``
        where `n` is the number of samples.
    """
    extra_args = {
        "lag": (lag, float),
        "phi": (phi, float),
        "sigma": (sigma, float),
    }
    check_in = _CheckInputs(n)
    check_in(**extra_args)

    epsilons = np.random.normal(0, sigma, n)
    etas = np.random.normal(0, sigma, n)

    x = np.zeros(n)
    y = etas

    for t in range(lag, n):
        x[t] = phi * epsilons[t] * y[t - lag]

    return x, y


def extinct_gaussian_process(n, phi=0.2, extinction_rate=0.5, radius=1):
    r"""
    2 nonlinearly dependent time series simulation.

    :math:`X_t` and :math:`Y_t` are together a bivariate nonlinear process.
    Noise :math:`\epsilon_t, \eta_t` are :math:`\mathcal{N}(0, \sigma)`, but chosen
    such that given radius, :math:`r`, uniform variable :math:`d \sim \mathcal{U}(0, 1)`,
    and extinction rate, :math:`p`:

    .. math::
        \begin{align*}
        \epsilon_t^2 + \eta_t^2 &> r^2\\
        d &> p
        \end{align*}

    With :math:`\phi = 0.2`,  :math:`:math:`X_t` and :math:`Y_t` follows:

    .. math::
        \begin{bmatrix} X_t \\ Y_t \end{bmatrix} =
        \begin{bmatrix} \phi & 0 \\ 0 & \phi \end{bmatrix}
        \begin{bmatrix} X_{t - 1} \\ Y_{t - 1} \end{bmatrix} +
        \begin{bmatrix} \epsilon_t \\ \eta_t \end{bmatrix}

    Parameters
    ----------
    n : int
        The number of samples desired by the simulation (>= 3).
    phi : float, default: 0.2
        The variance of the noise.
    extinction_rate : float, default: 0.5
        The rate at which the Gaussian process is extinct.
    radius : float, default: 1
        The radius of the Gaussian process.
    """
    extra_args = {
        "extinction_rate": (extinction_rate, float),
        "radius": (radius, float),
        "phi": (phi, float),
    }
    check_in = _CheckInputs(n)
    check_in(**extra_args)

    if (extinction_rate < 0) or (extinction_rate > 1):
        msg = "extinction_rate must be between 0 and 1, inclusive."
        raise ValueError(msg)

    x = np.zeros(n)
    y = np.zeros(n)

    for t in range(0, n):
        while True:
            eta = np.random.normal(0, 1)
            epsilon = np.random.normal(0, 1)
            d = np.random.uniform()

            if (eta**2 + epsilon**2 > radius**2) or (d > extinction_rate):
                break

        if t == 0:
            x[t] = eta
            y[t] = epsilon
        else:
            x[t] = phi * x[t - 1] + epsilon
            y[t] = phi * y[t - 1] + eta

    return x, y


TS_SIMS = {
    "indep_ar": indep_ar,
    "cross_corr_ar": cross_corr_ar,
    "nonlinear_process": nonlinear_process,
    "extinct_gaussian_process": extinct_gaussian_process,
}


def ts_sim(sim, n, **kwargs):
    r"""
    Time-series simulation generator.

    Takes a simulation and the required parameters, and outputs the simulated
    data matrices.

    Parameters
    ----------
    sim : str
        The name of the simulation (from the :mod:`hyppo.tools` module) that is to be
        rotated.
    n : int
        The number of samples desired by the simulation (>= 3).
    **kwargs
        Additional keyword arguements for the desired simulation.

    Returns
    -------
    x,y : ndarray of float
        Simulated data matrices.
    """
    if sim not in TS_SIMS.keys():
        raise ValueError(
            "sim_name must be one of the following: {}".format(list(TS_SIMS.keys()))
        )
    else:
        sim = TS_SIMS[sim]

    x, y = sim(n, **kwargs)

    if x.ndim == 1:
        x = x[:, np.newaxis]
    if y.ndim == 1:
        y = y[:, np.newaxis]

    return x, y



from ._utils import _CheckInputs
from .base import KSampleTest, KSampleTestOutput
from .ksamp import KSample


class MMD(KSampleTest):
    r"""
    Maximum Mean Discrepency (MMD) test statistic and p-value.

    MMD is a powerful multivariate 2-sample test. It leverages kernel similarity
    matrices
    capabilities (similar to tests like distance correlation or Dcorr). In fact, MMD
    statistic is equivalent to our 2-sample formulation nonparametric MANOVA via
    independence testing, i.e. :class:`hyppo.ksample.KSample`,
    and to
    :class:`hyppo.independence.Dcorr`,
    :class:`hyppo.ksample.DISCO`,
    :class:`hyppo.independence.Hsic`, and
    :class:`hyppo.ksample.Energy`
    :footcite:p:`pandaNonparMANOVAIndependence2021`
    :footcite:p:`shenExactEquivalenceDistance2020`.

    Parameters
    ----------
    compute_kernel : str, callable, or None, default: "gaussian"
        A function that computes the kernel similarity among the samples within each
        data matrix.
        Valid strings for ``compute_kernel`` are, as defined in
        :func:`sklearn.metrics.pairwise.pairwise_kernels`,

            [``"additive_chi2"``, ``"chi2"``, ``"linear"``, ``"poly"``,
            ``"polynomial"``, ``"rbf"``,
            ``"laplacian"``, ``"sigmoid"``, ``"cosine"``]

        Note ``"rbf"`` and ``"gaussian"`` are the same metric.
        Set to ``None`` or ``"precomputed"`` if ``x`` and ``y`` are already similarity
        matrices. To call a custom function, either create the similarity matrix
        before-hand or create a function of the form :func:`metric(x, **kwargs)`
        where ``x`` is the data matrix for which pairwise kernel similarity matrices are
        calculated and kwargs are extra arguements to send to your custom function.
    bias : bool, default: False
        Whether or not to use the biased or unbiased test statistics.
    **kwargs
        Arbitrary keyword arguments for ``compute_kernel``.

    Notes
    -----
    Traditionally, the formulation for the 2-sample MMD statistic
    is as follows :footcite:p:`grettonKernelTwoSampleTest2012`:

    Define
    :math:`\{ u_i \stackrel{iid}{\sim} F_U,\ i = 1, ..., n \}` and
    :math:`\{ v_j \stackrel{iid}{\sim} F_V,\ j = 1, ..., m \}` as two groups
    of samples deriving from different distributions with the same
    dimensionality. If :math:`k(\cdot, \cdot)` is a kernel metric (i.e. Gaussian)
    then,

    .. math::

        \mathrm{MMD}_{n, m}(\mathbf{u}, \mathbf{v}) =
        \frac{1}{m(m - 1)} \sum_{i = 1}^m \sum_{j \neq i}^m k(u_i, u_j)
        + \frac{1}{n(n - 1)} \sum_{i = 1}^n \sum_{j \neq i}^n k(v_i, v_j)
        - \frac{2}{mn} \sum_{i = 1}^n \sum_{j \neq i}^n k(v_i, v_j)

    The implementation in the :class:`hyppo.ksample.KSample` class (using
    :class:`hyppo.independence.Hsic` using 2 samples) is in
    fact equivalent to this implementation (for p-values) and statistics are
    equivalent up to a scaling factor :footcite:p:`shenExactEquivalenceDistance2020`.

    The p-value returned is calculated using a permutation test uses
    :meth:`hyppo.tools.perm_test`.
    The fast version of the test uses :meth:`hyppo.tools.chi2_approx`.

    References
    ----------
    .. footbibliography::
    """

    def __init__(self, compute_kernel="gaussian", bias=False, **kwargs):
        self.compute_kernel = compute_kernel
        self.bias = bias
        KSampleTest.__init__(self, compute_distance=None, bias=bias, **kwargs)

    def statistic(self, x, y):
        r"""
        Calulates the MMD test statistic.

        Parameters
        ----------
        x,y : ndarray of float
            Input data matrices. ``x`` and ``y`` must have the same number of
            dimensions. That is, the shapes must be ``(n, p)`` and ``(m, p)`` where
            `n` is the number of samples and `p` and `q` are the number of
            dimensions.

        Returns
        -------
        stat : float
            The computed MMD statistic.
        """
        n = x.shape[0]
        m = y.shape[0]

        # exact equivalence transformation Hsic and MMD
        stat = (
            KSample(
                indep_test="Hsic",
                compute_distkern=self.compute_kernel,
                bias=self.bias,
                **self.kwargs
            ).statistic(
                x,
                y,
            )
            * (2 * (n**2) * (m**2))
            / ((n + m) ** 4)
        )
        self.stat = stat

        return stat

    def test(self, x, y, reps=1000, workers=1, auto=True, random_state=None):
        r"""
        Calculates the MMD test statistic and p-value.

        Parameters
        ----------
        x,y : ndarray of float
            Input data matrices. ``x`` and ``y`` must have the same number of
            dimensions. That is, the shapes must be ``(n, p)`` and ``(m, p)`` where
            `n` is the number of samples and `p` and `q` are the number of
            dimensions.
        reps : int, default: 1000
            The number of replications used to estimate the null distribution
            when using the permutation test used to calculate the p-value.
        workers : int, default: 1
            The number of cores to parallelize the p-value computation over.
            Supply ``-1`` to use all cores available to the Process.
        auto : bool, default: True
            Automatically uses fast approximation when `n` and size of array
            is greater than 20. If ``True``, and sample size is greater than 20, then
            :class:`hyppo.tools.chi2_approx` will be run. Parameters ``reps`` and
            ``workers`` are
            irrelevant in this case. Otherwise, :class:`hyppo.tools.perm_test` will be
            run.

        Returns
        -------
        stat : float
            The computed MMD statistic.
        pvalue : float
            The computed MMD p-value.

        Examples
        --------
        >>> import numpy as np
        >>> from hyppo.ksample import MMD
        >>> x = np.arange(7)
        >>> y = x
        >>> stat, pvalue = MMD().test(x, y)
        >>> '%.3f, %.1f' % (stat, pvalue)
        '0.000, 1.0'
        """
        check_input = _CheckInputs(
            inputs=[x, y],
            indep_test="hsic",
        )
        x, y = check_input()

        # observed statistic
        stat = self.statistic(x, y)

        # since stat transformation is invariant under permutation, 2-sample Hsic
        # pvalue is identical to MMD
        _, pvalue = KSample(
            indep_test="Hsic",
            compute_distkern=self.compute_kernel,
            bias=self.bias,
            **self.kwargs
        ).test(x, y, reps=reps, workers=workers, auto=auto, random_state=random_state)

        return KSampleTestOutput(stat, pvalue)



import numpy as np
import pytest
from numpy.testing import assert_almost_equal
from sklearn.metrics import pairwise_distances

from ...tools import linear, power
from .. import HHG


class TestHHGStat:
    @pytest.mark.parametrize(
        "n, obs_stat", [(10, 560.0), (50, 112800.0), (100, 950600.0)]
    )
    @pytest.mark.parametrize("obs_pvalue", [1 / 1000])
    def test_linear_oned(self, n, obs_stat, obs_pvalue):
        np.random.seed(123456789)
        x, y = linear(n, 1)
        stat, pvalue = HHG().test(x, y)

        assert_almost_equal(stat, obs_stat, decimal=2)
        assert_almost_equal(pvalue, obs_pvalue, decimal=2)

    def test_diststat(self):
        np.random.seed(123456789)
        x, y = linear(100, 1)
        distx = pairwise_distances(x, x)
        disty = pairwise_distances(y, y)
        stat = HHG().statistic(distx, disty)

        assert_almost_equal(stat, 950600.0, decimal=2)

    @pytest.mark.parametrize("n", [10, 200])
    def test_rep(self, n):
        x, y = linear(n, 1)
        stat, pvalue = HHG().test(x, y, random_state=2)
        stat2, pvalue2 = HHG().test(x, y, random_state=2)

        assert stat == stat2
        assert pvalue == pvalue2


class TestHHGTypeIError:
    def test_oned(self):
        np.random.seed(123456789)
        est_power = power(
            "HHG",
            sim_type="indep",
            sim="multimodal_independence",
            n=50,
            p=1,
            alpha=0.05,
        )

        assert_almost_equal(est_power, 0.05, decimal=2)



import numpy as np
import pytest
from numpy.testing import (
    assert_allclose,
    assert_almost_equal,
    assert_array_equal,
    assert_array_less,
    assert_equal,
    assert_raises,
    assert_warns,
)
from sklearn.metrics import pairwise_distances
from sklearn.metrics.pairwise import pairwise_kernels

from ...independence import Dcorr
from ..common import (
    _check_distmat,
    _check_kernmat,
    _multi_check_kernmat,
    _PermTree,
    check_ndarray_xy,
    check_perm_blocks,
    check_perm_blocks_dim,
    check_reps,
    chi2_approx,
    compute_dist,
    compute_kern,
    contains_nan,
    convert_xy_float64,
    perm_test,
    multi_compute_kern,
    multi_perm_test,
)
from ..indep_sim import linear


class TestPermTree:
    """
    Tests that permutations are restricted correctly
    """

    def test_within_permutatins(self):
        np.random.seed(1)
        # i.e. case: y = np.asarray([0,1,0,1,0,1])
        blocks = np.vstack(
            (
                [-1, 1],
                [-1, 2],
                [-2, 1],
                [-2, 2],
                [-3, 1],
                [-3, 2],
            )
        )
        perm_tree = _PermTree(blocks)
        original_indices = perm_tree.original_indices()
        perms = np.asarray([perm_tree.permute_indices() for _ in range(10)])
        assert_array_less(np.abs(original_indices - perms), 2)
        assert_allclose(
            np.mean(perms, axis=0), [0.5, 0.5, 2.5, 2.5, 4.5, 4.5], rtol=0, atol=0.2
        )

    def test_across_permutations(self):
        np.random.seed(0)
        # i.e. case: y = np.asarray([0,0,1,1,2,2])
        blocks = np.vstack(
            (
                [1, -1],
                [1, -2],
                [2, -1],
                [2, -2],
                [3, -1],
                [3, -2],
            )
        )
        perm_tree = _PermTree(blocks)
        _ = perm_tree.original_indices()
        perms = np.asarray([perm_tree.permute_indices() for _ in range(100)])
        assert_equal(perms[0][1::2] - perms[0][::2], [1, 1, 1])
        assert_allclose(np.mean(perms, axis=0), [2, 3, 2, 3, 2, 3], rtol=0, atol=0.2)

    def test_fixed_permutation(self):
        np.random.seed(0)
        blocks = [-1, -2, -3, -4]
        perm_tree = _PermTree(blocks)
        assert_equal(perm_tree.permute_indices(), perm_tree.original_indices())

    def test_semifixed(self):
        np.random.seed(1)
        blocks = [1, 2, -3, -4]
        perm_tree = _PermTree(blocks)
        perms = np.asarray([perm_tree.permute_indices() for _ in range(10)])
        assert_equal(perms[0][2:], perm_tree.original_indices()[2:])
        assert_allclose(np.mean(perms, axis=0)[:2], [0.5, 0.5], rtol=0, atol=0.2)

    def test_non_int_inputs(self):
        blocks = ["a", "b", "c"]
        _ = _PermTree(blocks)

    def test_block_sizes(self):
        blocks = [1, 1, 2]
        assert_raises(ValueError, _PermTree, blocks)

    def test_noblock(self):
        perm_blocks = check_perm_blocks(None)
        assert_equal(perm_blocks, None)

    def test_not_ndarray(self):
        perm_blocks = (1, 2, 3)
        assert_raises(TypeError, check_perm_blocks, perm_blocks)

    def test_incorrect_dim(self):
        perm_blocks = np.arange(125).reshape(5, 5, 5)
        assert_raises(ValueError, check_perm_blocks, perm_blocks)

    def test_permblock_dim(self):
        perm_blocks = np.arange(100).reshape(50, 2)
        y = np.arange(100).reshape(10, 10)
        assert_raises(ValueError, check_perm_blocks_dim, perm_blocks, y)


class TestErrorWarn:
    """Tests errors and warnings."""

    def test_error_notndarray(self):
        # raises error if x or y is not a ndarray
        x = np.arange(20)
        y = [5] * 20
        assert_raises(TypeError, check_ndarray_xy, x, y)
        assert_raises(TypeError, check_ndarray_xy, y, x)

    def test_conv_float(self):
        # raises error if number of samples different (n)
        x = np.arange(20)
        y = np.arange(20)
        x, y = convert_xy_float64(x, y)
        assert_equal(x.dtype, np.float64)
        assert_equal(y.dtype, np.float64)

    def test_error_distkern(self):
        # raises error if samples are low (< 3)
        x = np.arange(10).reshape(-1, 1)
        y = np.arange(10).reshape(1, -1)
        assert_raises(ValueError, _check_distmat, x, y)
        assert_raises(ValueError, _check_kernmat, x, y)

    def test_error_threedist(self):
        x = np.arange(10).reshape(-1, 1)
        y = np.arange(10).reshape(1, -1)
        z = np.arange(10).reshape(1, -1)
        assert_raises(ValueError, _check_distmat, x, y, z)

    def test_error_nonzerodiag(self):
        x = np.eye(10, dtype=int) ^ 1  # 0 on diag, 1 elsewhere
        y = np.eye(10, dtype=int) ^ 1  # 0 on diag, 1 elsewhere
        z = np.eye(10, dtype=int)

        assert_raises(ValueError, _check_distmat, x, y, z)

    def test_error_multidistkern(self):
        # raises error if samples are low (< 3)
        x = np.arange(10).reshape(-1, 1)
        y = np.arange(10).reshape(1, -1)
        assert_raises(ValueError, _multi_check_kernmat, x, y)

    def test_error_nans(self):
        # raises error if inputs contain NaNs
        x = np.arange(20, dtype=float)
        x[0] = np.nan
        assert_raises(ValueError, contains_nan, x)

    @pytest.mark.parametrize(
        "reps", [-1, "1"]  # reps is negative  # reps is not integer
    )
    def test_error_reps(self, reps):
        # raises error if reps is negative
        assert_raises(ValueError, check_reps, reps)

    def test_warn_reps(self):
        # raises error if reps is negative
        reps = 100
        assert_warns(RuntimeWarning, check_reps, reps)


class TestHelper:
    """Tests errors and warnings derived from MGC."""

    def test_diskern(self):
        np.random.seed(123456789)
        x, y = linear(100, 1)
        distx = pairwise_distances(x, x)
        disty = pairwise_distances(y, y)

        l1 = pairwise_distances(x, metric="l1")
        n = l1.shape[0]
        med = np.median(
            np.lib.stride_tricks.as_strided(
                l1, (n - 1, n + 1), (l1.itemsize * (n + 1), l1.itemsize)
            )[:, 1:]
        )
        gamma = 1.0 / (2 * (med**2))

        kernx = pairwise_kernels(x, x, metric="rbf", gamma=gamma)
        kerny = pairwise_kernels(y, y, metric="rbf", gamma=gamma)

        distx, disty = compute_dist(distx, disty, metric=None)
        kernx, kerny = compute_kern(kernx, kerny, metric=None)
        distx_comp, disty_comp = compute_dist(x, y)
        kernx_comp, kerny_comp = compute_kern(x, y)
        kernx_comp1, kerny_comp1 = compute_kern(x, y, metric="rbf")

        assert_array_equal(distx, distx_comp)
        assert_array_equal(disty, disty_comp)
        assert_array_equal(kernx, kernx_comp)
        assert_array_equal(kerny, kerny_comp)
        assert_array_equal(kernx, kerny_comp1)
        assert_array_equal(kerny, kerny_comp1)

        def gaussian(x, **kwargs):
            return pairwise_kernels(x, x, metric="rbf", **kwargs)

        def euclidean(x, **kwargs):
            return pairwise_distances(x, x, metric="euclidean", **kwargs)

        distx, disty = compute_dist(x, y, metric=euclidean)
        kernx, kerny = compute_kern(x, y, metric=gaussian, gamma=gamma)

        assert_array_equal(distx, distx_comp)
        assert_array_equal(disty, disty_comp)
        assert_array_equal(kernx, kernx_comp)
        assert_array_equal(kerny, kerny_comp)

    def test_multidiskern(self):
        np.random.seed(123456789)
        x, y = linear(100, 1)
        distx = pairwise_distances(x, x)
        disty = pairwise_distances(y, y)

        l1 = pairwise_distances(x, metric="l1")
        n = l1.shape[0]
        med = np.median(
            np.lib.stride_tricks.as_strided(
                l1, (n - 1, n + 1), (l1.itemsize * (n + 1), l1.itemsize)
            )[:, 1:]
        )
        gamma = 1.0 / (2 * (med**2))

        kernx = pairwise_kernels(x, x, metric="rbf", gamma=gamma)
        kerny = pairwise_kernels(y, y, metric="rbf", gamma=gamma)

        distx, disty = compute_dist(distx, disty, metric=None)
        kernx, kerny = multi_compute_kern(*(kernx, kerny), metric=None)
        distx_comp, disty_comp = compute_dist(x, y)
        kernx_comp, kerny_comp = multi_compute_kern(*(x, y))
        kernx_comp1, kerny_comp1 = multi_compute_kern(*(x, y), metric="rbf")

        assert_array_equal(distx, distx_comp)
        assert_array_equal(disty, disty_comp)
        assert_array_equal(kernx, kernx_comp)
        assert_array_equal(kerny, kerny_comp)
        assert_array_equal(kernx, kerny_comp1)
        assert_array_equal(kerny, kerny_comp1)

        def gaussian(x, **kwargs):
            return pairwise_kernels(x, x, metric="rbf", **kwargs)

        def euclidean(x, **kwargs):
            return pairwise_distances(x, x, metric="euclidean", **kwargs)

        distx, disty = compute_dist(x, y, metric=euclidean)
        kernx, kerny = multi_compute_kern(*(x, y), metric=gaussian, gamma=gamma)

        assert_array_equal(distx, distx_comp)
        assert_array_equal(disty, disty_comp)
        assert_array_equal(kernx, kernx_comp)
        assert_array_equal(kerny, kerny_comp)

    def test_permtest(self):
        x, y = linear(100, 1)

        stat, pvalue, _ = perm_test(Dcorr().statistic, x, y, is_distsim=False)
        assert_almost_equal(stat, 1.0, decimal=1)
        assert_almost_equal(pvalue, 1 / 1000, decimal=1)

        x = pairwise_distances(x, x)
        y = pairwise_distances(y, y)
        stat, pvalue, _ = perm_test(Dcorr().statistic, x, y, is_distsim=True)
        assert_almost_equal(stat, 1.0, decimal=1)
        assert_almost_equal(pvalue, 1 / 1000, decimal=1)

    def test_multipermtest(self):
        x, y = linear(100, 1)

        stat, pvalue, _ = multi_perm_test(Dcorr().statistic, *(x, y))
        assert_almost_equal(stat, 1.0, decimal=1)
        assert_almost_equal(pvalue, 1 / 1000, decimal=1)

    def test_chi2(self):
        x, y = linear(100, 1)

        stat, pvalue = chi2_approx(Dcorr().statistic, x, y)
        assert_almost_equal(stat, 1.0, decimal=1)
        assert_almost_equal(pvalue, 1 / 1000, decimal=1)



from .cdcorr import ConditionalDcorr
from .FCIT import FCIT
from .kci import KCI
from .pcorr import PartialCorr
from .pdcorr import PartialDcorr

__all__ = [s for s in dir()]  # add imported tests to __all__

COND_INDEP_TESTS = {
    "fcit": FCIT,
    "kci": KCI,
    "conditionaldcorr": ConditionalDcorr,
    "partialcorr": PartialCorr,
    "partialdcorr": PartialDcorr,
}



import numpy as np

from ._utils import _CheckInputs
from .base import IndependenceTest


class CCA(IndependenceTest):
    r"""
    Cannonical Correlation Analysis (CCA) test statistic and p-value.

    This test can be thought of inferring information from cross-covariance
    matrices :footcite:p:`hardleCanonicalCorrelationAnalysis2015`.
    It has been thought that virtually all parametric tests
    of significance can be treated as a special case of CCA
    :footcite:p:`knappCanonicalCorrelationAnalysis1978`. The
    method was first introduced by :footcite:t:`hotellingRelationsTwoSets1992`.

    Notes
    -----
    The statistic can be derived as follows
    :footcite:p:`hardoonCanonicalCorrelationAnalysis2004`:

    Let :math:`x` and :math:`y` be :math:`(n, p)` samples of random variables
    :math:`X` and :math:`Y`. We can center :math:`x` and :math:`y` and then
    calculate the sample covariance matrix :math:`\hat{\Sigma}_{xy} = x^T y`
    and the variance matrices for :math:`x` and :math:`y` are defined
    similarly. Then, the CCA test statistic is found by calculating vectors
    :math:`a \in \mathbb{R}^p` and :math:`b \in \mathbb{R}^q` that maximize

    .. math::

        \mathrm{CCA}_n (x, y) =
            \max_{a \in \mathbb{R}^p, b \in \mathbb{R}^q}
            \frac{a^T \hat{\Sigma}_{xy} b}
            {\sqrt{a^T \hat{\Sigma}_{xx} a}
             \sqrt{b^T \hat{\Sigma}_{yy} b}}

    The p-value returned is calculated using a permutation test using
    :meth:`hyppo.tools.perm_test`.

    References
    ----------
    .. footbibliography::
    """

    def __init__(self):
        IndependenceTest.__init__(self)

    def statistic(self, x, y):
        r"""
        Helper function that calculates the CCA test statistic.

        Parameters
        ----------
        x,y : ndarray of float
            Input data matrices. ``x`` and ``y`` must have the same number of
            samples and dimensions. That is, the shapes must be ``(n, p)`` where
            `n` is the number of samples and `p` is the number of dimensions.

        Returns
        -------
        stat : float
            The computed CCA statistic.
        """
        # center each matrix
        centx = x - np.mean(x, axis=0)
        centy = y - np.mean(y, axis=0)

        # calculate covariance and variances for inputs
        covar = centx.T @ centy
        varx = centx.T @ centx
        vary = centy.T @ centy

        # if 1-d, don't calculate the svd
        if varx.size == 1 or vary.size == 1 or covar.size == 1:
            covar = np.sum(covar**2)
            stat = covar / np.sqrt(np.sum(varx**2) * np.sum(vary**2))
        else:
            covar = np.sum(np.linalg.svd(covar, 1)[1] ** 2)
            stat = covar / np.sqrt(
                np.sum(np.linalg.svd(varx, 1)[1] ** 2)
                * np.sum(np.linalg.svd(vary, 1)[1] ** 2)
            )
        self.stat = stat

        return stat

    def test(self, x, y, reps=1000, workers=1, random_state=None):
        r"""
        Calculates the CCA test statistic and p-value.

        Parameters
        ----------
        x,y : ndarray of float
            Input data matrices. ``x`` and ``y`` must have the same number of
            samples and dimensions. That is, the shapes must be ``(n, p)`` where
            `n` is the number of samples and `p` is the number of dimensions.
        reps : int, default: 1000
            The number of replications used to estimate the null distribution
            when using the permutation test used to calculate the p-value.
        workers : int, default: 1
            The number of cores to parallelize the p-value computation over.
            Supply ``-1`` to use all cores available to the Process.

        Returns
        -------
        stat : float
            The computed CCA statistic.
        pvalue : float
            The computed CCA p-value.

        Examples
        --------
        >>> import numpy as np
        >>> from hyppo.independence import CCA
        >>> x = np.arange(7)
        >>> y = x
        >>> stat, pvalue = CCA().test(x, y)
        >>> '%.1f, %.2f' % (stat, pvalue)
        '1.0, 0.00'
        """
        check_input = _CheckInputs(x, y, reps=reps)
        x, y = check_input()

        # use default permutation test
        return super(CCA, self).test(
            x, y, reps, workers, is_distsim=False, random_state=random_state
        )



import time

import joblib
import numpy as np
from scipy.stats import ttest_1samp
from sklearn.metrics import mean_squared_error as mse
from sklearn.model_selection import GridSearchCV, ShuffleSplit
from sklearn.preprocessing import StandardScaler
from sklearn.tree import DecisionTreeRegressor

from .base import ConditionalIndependenceTest, ConditionalIndependenceTestOutput


class FCIT(ConditionalIndependenceTest):
    r"""
    Fast Conditional Independence test statistic and p-value

    The Fast Conditional Independence Test is a non-parametric
    conditional independence test :footcite:p:`chalupka2018FastConditionalIndependence`.

    Parameters
    ----------
    model: Sklearn regressor
        Regressor used to predict input data :math:`Y`
    cv_grid: dict
       Dictionary of parameters to cross-validate over when training regressor.
    num_perm: int
        Number of data permutations to estimate the p-value from marginal stats.
    prop_test: float
        Proportion of data to evaluate test stat on.
    discrete: tuple of string
        Whether :math:`X` or :math:`Y` are discrete

    Notes
    -----
    .. note::
       This algorithm is currently a pre-print on arXiv.

    The motivation for the test rests on the assumption that if :math:`X \not\!\perp\!\!\!\perp Y \mid Z`,
    then :math:`Y` should be more accurately predicted by using both
    :math:`X` and :math:`Z` as covariates as opposed to only using
    :math:`Z` as a covariate. Likewise, if :math:`X \perp \!\!\! \perp Y \mid Z`,
    then :math:`Y` should be predicted just as accurately by solely
    using :math:`X` or soley using :math:`Z` :footcite:p:`chalupka2018FastConditionalIndependence`.
    Thus, the test works by using a regressor (the default is decision tree) to
    to predict input :math:`Y` using both :math:`X` and :math:`Z` and using
    only :math:`Z` :footcite:p:`chalupka2018FastConditionalIndependence`. Then,
    accuracy of both predictions are measured via mean-squared error (MSE).
    :math:`X \perp \!\!\! \perp Y \mid Z` if and only if MSE of the algorithm
    using both :math:`X` and :math:`Z` is not smaller than the MSE of the
    algorithm trained using only :math:`Z` :footcite:p:`chalupka2018FastConditionalIndependence`.

    References
    ----------
    .. footbibliography::
    """

    def __init__(
        self,
        model=DecisionTreeRegressor(),
        cv_grid={"min_samples_split": [2, 8, 64, 512, 1e-2, 0.2, 0.4]},
        num_perm=8,
        prop_test=0.1,
        discrete=(False, False),
    ):

        self.model = model
        self.cv_grid = cv_grid
        self.num_perm = num_perm
        self.prop_test = prop_test
        self.discrete = discrete
        ConditionalIndependenceTest.__init__(self)

    def statistic(self, x, y, z=None):
        r"""
        Calculates the FCIT test statistic.

        Parameters
        ----------
        x,y,z : ndarray of float
            Input data matrices.

        Returns
        -------
        stat : float
            The computed FCIT test statistic.
        two_sided: float
            Two-sided p-value associated with test statistic
        """

        n_samples = x.shape[0]
        n_test = int(n_samples * self.prop_test)

        data_permutations = [
            np.random.permutation(x.shape[0]) for i in range(self.num_perm)
        ]

        clf = _cross_val(x, y, z, self.cv_grid, self.model, prop_test=self.prop_test)
        datadict = {
            "x": x,
            "y": y,
            "z": z,
            "data_permutation": data_permutations,
            "n_test": n_test,
            "reshuffle": False,
            "clf": clf,
        }
        d1_stats = np.array(
            joblib.Parallel(n_jobs=-1, max_nbytes=100e6)(
                joblib.delayed(_obtain_error)((datadict, i))
                for i in range(self.num_perm)
            )
        )

        if z.shape[1] == 0:
            x_indep_y = x[np.random.permutation(n_samples)]
        else:
            x_indep_y = np.empty([x.shape[0], 0])

        clf = _cross_val(
            x_indep_y, y, z, self.cv_grid, self.model, prop_test=self.prop_test
        )

        datadict["reshuffle"] = True
        datadict["x"] = x_indep_y
        d0_stats = np.array(
            joblib.Parallel(n_jobs=-1, max_nbytes=100e6)(
                joblib.delayed(_obtain_error)((datadict, i))
                for i in range(self.num_perm)
            )
        )

        stat, two_sided = ttest_1samp(d0_stats / d1_stats, 1)

        return stat, two_sided

    def test(self, x, y, z=None):
        r"""
        Calculates the FCIT test statistic and p-value.

        Parameters
        ----------
        x,y,z : ndarray of float
            Input data matrices.

        Returns
        -------
        stat : float
            The computed FCIT statistic.
        pvalue : float
            The computed FCIT p-value.

        Examples
        --------
        >>> import numpy as np
        >>> from hyppo.conditional import FCIT
        >>> from sklearn.tree import DecisionTreeRegressor
        >>> np.random.seed(1234)
        >>> dim = 2
        >>> n = 100000
        >>> z1 = np.random.multivariate_normal(mean=np.zeros(dim), cov=np.eye(dim), size=(n))
        >>> A1 = np.random.normal(loc=0, scale=1, size=dim * dim).reshape(dim, dim)
        >>> B1 = np.random.normal(loc=0, scale=1, size=dim * dim).reshape(dim, dim)
        >>> x1 = (A1 @ z1.T + np.random.multivariate_normal(mean=np.zeros(dim), cov=np.eye(dim), size=(n)).T)
        >>> y1 = (B1 @ z1.T + np.random.multivariate_normal(mean=np.zeros(dim), cov=np.eye(dim), size=(n)).T)
        >>> model = DecisionTreeRegressor()
        >>> cv_grid = {"min_samples_split": [2, 8, 64, 512, 1e-2, 0.2, 0.4]}
        >>> stat, pvalue = FCIT(model=model, cv_grid=cv_grid).test(x1.T, y1.T, z1)
        >>> '%.1f, %.2f' % (stat, pvalue)
        '-3.6, 1.00'
        """

        n_samples = x.shape[0]

        if z is None:
            z = np.empty([n_samples, 0])

        if self.discrete[0] and not self.discrete[1]:
            x, y = y, x
        elif x.shape[1] < y.shape[1]:
            x, y = y, x

        y = StandardScaler().fit_transform(y)

        stat, two_sided = self.statistic(x, y, z)

        if stat < 0:
            pvalue = 1 - two_sided / 2
        else:
            pvalue = two_sided / 2

        return ConditionalIndependenceTestOutput(stat, pvalue)


def _cross_val(x, y, z, cv_grid, model, prop_test):
    """
    Choose the regression hyperparameters by
    cross-validation.
    """

    splitter = ShuffleSplit(n_splits=3, test_size=prop_test)
    cv = GridSearchCV(estimator=model, cv=splitter, param_grid=cv_grid, n_jobs=-1)
    cv.fit(_interleave(x, z), y)

    return type(model)(**cv.best_params_)


def _interleave(x, z, seed=None):
    """Interleave x and z dimension-wise."""
    state = np.random.get_state()
    np.random.seed(seed or int(time.time()))
    total_ids = np.random.permutation(x.shape[1] + z.shape[1])
    np.random.set_state(state)
    out = np.zeros([x.shape[0], x.shape[1] + z.shape[1]])
    out[:, total_ids[: x.shape[1]]] = x
    out[:, total_ids[x.shape[1] :]] = z
    return out


def _obtain_error(data_and_i):
    """
    A function used for multithreaded computation of the fcit test statistic.
    Calculates MSE error for both trained regressors.
    """
    data, i = data_and_i
    x = data["x"]
    y = data["y"]
    z = data["z"]
    if data["reshuffle"]:
        perm_ids = np.random.permutation(x.shape[0])
    else:
        perm_ids = np.arange(x.shape[0])
    data_permutation = data["data_permutation"][i]
    n_test = data["n_test"]
    clf = data["clf"]

    x_z = _interleave(x[perm_ids], z, seed=i)

    clf.fit(x_z[data_permutation][n_test:], y[data_permutation][n_test:])
    return mse(
        y[data_permutation][:n_test], clf.predict(x_z[data_permutation][:n_test])
    )



from .common import *
from .conditional_indep_sim import *
from .indep_sim import *
from .ksample_sim import *
from .power import *
from .time_series_sim import *

__all__ = [s for s in dir() if not s.startswith("_")]  # remove dunders



import numpy as np
import pytest
from numpy.testing import assert_almost_equal, assert_raises, assert_warns

from .. import DiscrimOneSample


class TestOneSample:
    def test_same_one(self):
        # matches test calculated statistics and p-value for indiscriminable subjects
        x = np.ones((100, 2), dtype=float)
        y = np.concatenate((np.zeros(50), np.ones(50)), axis=0)

        np.random.seed(123456789)
        obs_stat = 0.5
        obs_p = 1.0
        stat, p, _ = DiscrimOneSample().test(x, y, reps=10)

        assert_almost_equal(stat, obs_stat, decimal=2)
        assert_almost_equal(p, obs_p, decimal=2)

    def test_diff_one(self):
        # matches test calculated statistics and p-value for discriminable subjects
        x = np.concatenate((np.zeros((50, 2)), np.ones((50, 2))), axis=0)
        y = np.concatenate((np.zeros(50), np.ones(50)), axis=0)

        np.random.seed(123456789)
        obs_stat = 1.0
        obs_p = 0.0909090909090909
        stat, p, _ = DiscrimOneSample().test(x, y, reps=10)

        assert_almost_equal(stat, obs_stat, decimal=3)
        assert_almost_equal(p, obs_p, decimal=3)


class TestOneSampleWarn:
    """Tests errors and warnings derived from one sample test."""

    def test_error_one_id(self):
        # checks whether y has only one id
        X = np.ones((100, 2), dtype=float)
        Y = np.ones((100, 1))

        assert_raises(ValueError, DiscrimOneSample().test, X, Y)

    def test_error_nans(self):
        # raises error if inputs contain NaNs
        x = np.arange(20, dtype=float)
        x[0] = np.nan
        assert_raises(ValueError, DiscrimOneSample().test, x, x)

        y = np.arange(20)
        assert_raises(ValueError, DiscrimOneSample().test, x, y)

    @pytest.mark.parametrize(
        "reps", [-1, "1"]  # reps is negative  # reps is not integer
    )
    def test_error_reps(self, reps):
        # raises error if reps is negative
        x = np.ones((100, 2), dtype=float)
        y = np.concatenate((np.zeros(50), np.ones(50)), axis=0)

        assert_raises(ValueError, DiscrimOneSample().test, x, y, reps=reps)

    def test_min_sample(self):
        # raises warning when sample number is less than 10
        x = np.ones((6, 2), dtype=float)
        y = np.concatenate((np.zeros(3), np.ones(3)), axis=0)

        assert_raises(ValueError, DiscrimOneSample().test, x, y)
